#! /usr/bin/python


from twisted.internet import gtk2reactor
gtk2reactor.install()
from twisted.internet import reactor
from twisted.python import log

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import agent, addressbook, message, transport, permission
from petmail.gui.ticketAgent import TicketAgent

import pygtk
pygtk.require("2.0")
import gtk, gtk.glade, gobject

import pickle, os, os.path, sys, time
#log.startLogging(sys.stderr)

def sibling(name):
    script = sys.argv[0]
    d = os.path.dirname(script)
    return os.path.join(d, name)

gladefile = sibling("gui.glade")

class Helper(object):

    def set_sensitive(self, names, sensitive):
        for name in names:
            self.w(name).set_sensitive(sensitive)

    def w(self, name):
        return self.xml.get_widget(name)
    def text(self, name):
        return self.w(name).get_text()

class LogWindow(Helper):
    def __init__(self, gui):
        xml = gtk.glade.XML(gladefile, "log")
        self.xml = xml
        self.parent = gui
        self.gui = gui
        self.model = gtk.TreeStore(str, str, object)
        xml.signal_connect("do_log_close", self.do_close)
        xml.signal_connect("do_log_clear", self.do_clear)
        xml.signal_connect("do_log_append", self.do_append)

        view = self.w("log_view")
        view.set_model(self.model)
        self.view = view

        col = gtk.TreeViewColumn("Time", gtk.CellRendererText(), text=0)
        col.set_resizable(1)
        view.append_column(col)

        col = gtk.TreeViewColumn("Event", gtk.CellRendererText(), text=1)
        col.set_resizable(1)
        view.append_column(col)

        view.get_selection().set_mode(gtk.SELECTION_NONE)
        view.connect("button_press_event", self.do_clicked)

        self.window = self.w("log")
        self.window.connect("destroy", self.do_destroy)
        self.fill()

    def stringify(self, message):
        return ' '.join(map(str, message))

    def render(self, event):
        timestamp = event.time
        #date = time.strftime("%d %b %Y", time.localtime(timestamp))
        t = time.strftime("%H:%M:%S", time.localtime(timestamp))
        text = self.stringify(event.getMessage())
        return (t, text, event)
        
    def fill(self):
        self.model.clear()
        l = self.gui.agent.eventLog
        for event in l.getItems(None):
            self.fill_entry(None, event)
        l.addObserver(self.add)
        self.scroll_to_bottom()

    def fill_entry(self, viewparent, lognode):
        fields = self.render(lognode)
        row = self.model.append(viewparent, fields)
        for child in lognode.getChildren():
            self.fill_entry(row, child)


    def add_helper(self, model, path, iter, data):
        parent, fields, found = data
        if model[path][2] == parent:
            model.append(iter, fields)
            # do we need to emit a signal here? apparently not.
            found[0] = True
            return True
        return False

    def add(self, event, pclass, parent):
        fields = self.render(event)
        if parent:
            # find the parent. It might be in a sub-node, so we must search
            # recursively. Ick. This will eventually be replaced with
            # something more efficient.
            found = [False]
            self.model.foreach(self.add_helper, (parent, fields, found))
            if not found[0]:
                print "HEY!, no parent", fields
                print " parent=", self.stringify(parent.getMessage())
            return
        self.model.append(None, fields)
        self.scroll_to_bottom()

    def scroll_to_bottom(self):
        row = self.model[len(self.model)-1]
        self.view.scroll_to_cell(row.path)

    def show(self):
        self.window.deiconify()
    def do_clicked(self, *args):
        #print "clicked", args
        pass

    def do_clear(self, widget):
        self.model.clear()

    def do_append(self, widget):
        self.gui.agent.log("adding fake line")

    def do_close(self, widget):
        self.window.destroy()
    def do_destroy(self, widget):
        self.gui.agent.eventLog.removeObserver(self.add)
        self.gui.event_log_closed()

class OwnerWindow(Helper):
    def __init__(self, mepane, owner):
        self.mepane = mepane
        self.gui = mepane.gui
        self.owner = owner

        xml = gtk.glade.XML(gladefile, "owner")
        self.xml = xml

        g = xml.get_widget
        self.window = g("owner")
        self.ok_button = g("my_ok_button")

        c = xml.signal_connect
        c("do_dirty", self.do_dirty)
        c("on_owner_inlet_maildir_toggled", self.inlet_maildir_toggled)
        c("on_owner_inlet_mailsocket_toggled", self.inlet_mailsocket_toggled)
        c("on_owner_inlet_pb_toggled", self.inlet_pb_toggled)
        c("on_owner_inlet_none_toggled", self.inlet_none_toggled)
        c("on_my_maxsize_toggle_toggled", self.maxsize_toggled)

        c("do_me_cancel", self.do_close)
        c("do_me_delete", self.do_delete)
        c("do_me_ok", self.do_save)

        self.window.connect("destroy", self.do_destroy)
        self.populate()


    def populate(self):
        owner = self.owner
        self.window.set_title("My Identity '%s'" % owner.getName())

        self.w("my_ID_label").set_text(owner.ID)

        myname = ""
        if owner.myName != None:
            myname = owner.myName
        self.w("my_local_name_entry").set_text(myname)

        names = ""
        if owner.names != None:
            names = owner.names
        self.w("my_name_entry").set_text(names)

        self.w("my_addresses_entry").set_text(", ".join(owner.addresses))

        if owner.inlets:
            inlet = owner.inlets[0]
        else:
            inlet = None
        if isinstance(inlet, transport.MaildirInlet):
            self.w("owner_inlet_maildir").set_active(1)
            self.w("owner_inlet_mailsocket").set_active(0)
            self.w("owner_inlet_pb").set_active(0)
            self.w("owner_inlet_none").set_active(0)
            self.w("owner_inlet_maildir_directory").set_text(inlet.basedir)
            self.w("owner_inlet_maildir_address").set_text(inlet.address)
        elif isinstance(inlet, transport.MailSocket):
            self.w("owner_inlet_maildir").set_active(0)
            self.w("owner_inlet_mailsocket").set_active(1)
            self.w("owner_inlet_pb").set_active(0)
            self.w("owner_inlet_none").set_active(0)
            port = str(inlet.portnum)
            self.w("owner_inlet_mailsocket_port").set_text(port)
            addr = inlet.address
            self.w("owner_inlet_mailsocket_address").set_text(addr)
        elif isinstance(inlet, transport.PBInlet):
            self.w("owner_inlet_maildir").set_active(0)
            self.w("owner_inlet_mailsocket").set_active(0)
            self.w("owner_inlet_pb").set_active(1)
            self.w("owner_inlet_none").set_active(0)
            self.w("owner_inlet_pb_hostname").set_text(inlet.hostname)
            self.w("owner_inlet_pb_port").set_text(str(inlet.portnum))
        else:
            self.w("owner_inlet_maildir").set_active(0)
            self.w("owner_inlet_mailsocket").set_active(0)
            self.w("owner_inlet_pb").set_active(0)
            self.w("owner_inlet_none").set_active(1)

        if owner.genericPermission.maxsize != None:
            self.w("my_maxsize_toggle").set_active(1)
            maxsize = float(owner.genericPermission.maxsize / 1000)
            self.w("my_maxsize_entry").set_value(maxsize)
            self.w("my_maxsize_entry").set_sensitive(1)
        else:
            self.w("my_maxsize_toggle").set_active(0)
            self.w("my_maxsize_entry").set_sensitive(0)

        self.w("my_accept_html").set_active(0)

        url_string = owner.genericPermission.getTicketTakersString()
        self.w("my_ticket_urls_entry").set_text(url_string)
                
        for n in ["owner_inlet_maildir", "owner_inlet_mailsocket",
                  "owner_inlet_pb", "owner_inlet_none"]:
            self.w(n).emit("toggled")

        self.ok_button.set_sensitive(0)

    def inlet_maildir_toggled(self, widget):
        fields = ["owner_inlet_maildir_directory_label",
                  "owner_inlet_maildir_directory",
                  "owner_inlet_maildir_address_label",
                  "owner_inlet_maildir_address"]
        self.set_sensitive(fields, widget.get_active())

    def inlet_mailsocket_toggled(self, widget):
        fields = ["owner_inlet_mailsocket_port_label",
                  "owner_inlet_mailsocket_port",
                  "owner_inlet_mailsocket_address_label",
                  "owner_inlet_mailsocket_address"]
        self.set_sensitive(fields, widget.get_active())

    def inlet_pb_toggled(self, widget):
        fields = ["owner_inlet_pb_hostname_label",
                  "owner_inlet_pb_hostname",
                  "owner_inlet_pb_port_label",
                  "owner_inlet_pb_port"]
        self.set_sensitive(fields, widget.get_active())

    def inlet_none_toggled(self, widget):
        fields = ["owner_inlet_none_label"]
        self.set_sensitive(fields, widget.get_active())

    def maxsize_toggled(self, widget):
        self.set_sensitive(["my_maxsize_entry"], widget.get_active())

    def do_close(self, widget):
        # close the window without making any changes
        self.window.destroy()
        #self.mepane.closeOwner(self.owner)

    def do_destroy(self, widget):
        # the window is being removed, we must let the MePane know
        self.mepane.closeOwner(self.owner)
        return False # do normal handling too

    def show(self):
        w = self.xml.get_widget("owner")
        w.deiconify() # this also appears to raise it to the top

    def do_dirty(self, *args):
        self.ok_button.set_sensitive(1)

    def do_delete(self, widget):
        # confirm
        xml = gtk.glade.XML(gladefile, "confirm_dialog")
        dialog = xml.get_widget("confirm_dialog")
        l = xml.get_widget("confirm_label")
        confirm = ""
        confirm += "All secret keys will be deleted.\n"
        confirm += "All old mail will be deleted.\n" # TODO: really?
        confirm += "All Permissions will be deleted.\n"
        confirm += "You will no longer be able to receive any mail"
        confirm += " addressed to this Identity in the future. Ever.\n"
        confirm += "(just to be sure, we will notify all current Permission"
        confirm += " holders and ask them to delete this entry from their"
        confirm += " addressbooks)\n"
        confirm += "\n"
        confirm += "<b>Really delete Identity '%s'?</b>\n" \
                   % self.owner.getName()
        l.set_markup(confirm)
        dialog.set_title("Delete Identity %s" % self.owner.ID)
        dialog.connect("response", self._confirm_response)
    def _confirm_response(self, dialog, responseid):
        if responseid == gtk.RESPONSE_OK:
            self.really_delete()
        dialog.destroy()
    def really_delete(self):
        self.gui.agent.removeOwner(self.owner)
        self.window.destroy()
        #self.mepane.closeOwner(self.owner)
        self.mepane.ownerModified(self.owner)

    def do_save(self, widget):
        print "saving"
        owner = self.owner

        # addresses
        text = self.w("my_addresses_entry").get_text()
        pieces = text.split(",")
        new = []
        for i in pieces:
            i = i.strip()
            if i:
                new.append(i)
        owner.addresses = new
        print new

        # local name
        myName = self.w("my_local_name_entry").get_text()
        myName.strip()
        owner.setLocalName(myName)

        # published name
        text = self.w("my_name_entry").get_text()
        text.strip()
        owner.setName(text)

        # inlet
        if self.w("owner_inlet_maildir").get_active():
            directory = self.w("owner_inlet_maildir_directory").get_text()
            address = self.w("owner_inlet_maildir_address").get_text()
            inlet = transport.MaildirInlet(directory, address)
        elif self.w("owner_inlet_mailsocket").get_active():
            portnum = int(self.w("owner_inlet_mailsocket_port").get_text())
            address = self.w("owner_inlet_mailsocket_address").get_text()
            inlet = transport.MailSocket(portnum, address)
        elif self.w("owner_inlet_pb").get_active():
            hostname = self.w("owner_inlet_pb_hostname").get_text()
            portnum = int(self.w("owner_inlet_pb_port").get_text())
            inlet = transport.PBInlet(hostname, portnum)
        elif self.w("owner_inlet_none").get_active():
            inlet = None
        else:
            log.msg("Weird, no owner_inlet selection is active")
        owner.setInlet(inlet)

        # our GenericPermission fields
        gp = owner.genericPermission
        if self.w("my_maxsize_toggle").get_active():
            maxsize = int(self.w("my_maxsize_entry").get_value()) * 1000
            gp.setMaxSize(maxsize)
        else:
            gp.setMaxSize(None)
        if self.w("my_accept_html").get_active():
            gp.setContentTypes(["*/*"])
        else:
            gp.setContentTypes(["text/plain"]) # TODO: needs work
        urlString = self.w("my_ticket_urls_entry").get_text()
        gp.setTicketTakersFromString(self.gui.agent, urlString)

        # save it all
        owner.publish(self.gui.agent.addressPublishAgents)
        self.gui.scheduleSave()

        self.populate() # make sure it was stored correctly
        self.mepane.ownerModified(self.owner)

class KeyGenerationProgress(object):
    def __init__(self, window, progress_widget, entropy_widget):
        self.window = window
        self.progress = progress_widget
        self.entropy = entropy_widget
        self.do_entropy = False
    def set_progress(self, progress):
        self.progress.set_fraction(progress)
    def pulse_progress(self):
        self.progress.pulse()
    def set_entropy(self, entropy):
        if not self.do_entropy:
            self.do_entropy = True
            self.entropy.show()
        self.entropy.set_fraction(entropy)
    def done(self):
        self.window.destroy()

class MePane(object):
    def __init__(self, gui):
        self.gui = gui
        self.ownerpanels = {}

    def load(self, xml):
        g = xml.get_widget
        xml.signal_connect("do_create_identity", self.do_create_identity)
        self.app = g('app')
        self.view = g('owner_tree')
        model = self.model = gtk.ListStore(gobject.TYPE_STRING,
                                           gobject.TYPE_STRING,
                                           gobject.TYPE_PYOBJECT)
        self.view.set_model(model)

        col = gtk.TreeViewColumn("Name", gtk.CellRendererText(), text=0)
        col.set_resizable(1)
        self.view.append_column(col)

        col = gtk.TreeViewColumn("ID", gtk.CellRendererText(), text=1)
        col.set_resizable(1)
        self.view.append_column(col)

        sel = self.view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        self.view.connect("button_press_event", self.do_clicked)


    def populate(self):
        self.model.clear()
        for owner in self.gui.agent.getOwners():
            i = self.model.append(None)
            self.model.set(i,
                           0, owner.getName(),
                           1, owner.ID,
                           2, owner)

    def do_clicked(self, widget, event):
        if event.type != gtk.gdk._2BUTTON_PRESS:
            return
        path, col, cx, cy = widget.get_path_at_pos(int(event.x),
                                                   int(event.y))
        sel = widget.get_selection()
        sel.select_path(path)
        model, iter = sel.get_selected()
        name = model.get_value(iter, 0)
        owner = model.get_value(iter, 2)
        self.inspectOwner(owner)

    def do_create_identity(self, widget):
        x = gtk.glade.XML(gladefile, "generateKey")
        p0 = x.get_widget("generateKey")
        p1 = x.get_widget("key_generation_progress")
        p2 = x.get_widget("entropy_level")
        p = KeyGenerationProgress(p0, p1, p2)
        
        d = self.gui.agent.createIdentity(p)
        d.addCallback(self._create1)
        d.addErrback(self._create_failed)

    def _create1(self, owner):
        self.gui.agent.addOwner(owner)
        self.ownerModified(owner)
        self.inspectOwner(owner)

    def _create_failed(self, why):
        m = gtk.MessageDialog(self.app, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                              "Unable to create Identity: %s" % why)
        m.connect("response", lambda m,y: m.destroy())
        m.show()

    def inspectOwner(self, owner):
        w = self.ownerpanels.get(owner)
        if not w:
            w = OwnerWindow(self, owner)
            self.ownerpanels[owner] = w
            w.populate()
        w.show() # makes sure it is on top, and not iconified

    def closeOwner(self, owner):
        # this is called after the OwnerWindow has been closed
        del self.ownerpanels[owner]

    def ownerModified(self, owner):
        # note! owner may no longer exist (we can be called in response to a
        # delete action)
        self.populate()
        # TODO: update selection
        self.gui.addrpane.populateOwners()
        self.gui.sendpane.populateOwners()

class AddressBookPane(Helper):
    def __init__(self, gui):
        self.gui = gui
        self.currentEntry = None
        self.owner = None

    def load(self, xml):
        self.xml = xml
        # addressbook_tree, ID_entry, names_entry, petname_entry, IMT_entry,
        # maxsize_entry, accept_html
        # outbound_IMT, outbound_maxsize, outbound_html
        view = self.addressbook_tree = xml.get_widget("addressbook_tree")
        model = self.addressbook_list = gtk.ListStore(gobject.TYPE_STRING,
                                                      gobject.TYPE_PYOBJECT)
        view.set_model(model)
        view.append_column(gtk.TreeViewColumn("Name",
                                              gtk.CellRendererText(),
                                              text=0))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_select)
        #self.populateAddressBook()
        g = xml.get_widget
        c = xml.signal_connect

        self.w("petname_entry").connect("activate",
                                        self.petname_entry_changed)

        c("do_lookup", self.do_lookup)
        c("on_maxsize_toggle_toggled", self.do_maxsize_toggled)
        c("on_IMT_toggle_toggled", self.do_IMG_toggled)
        c("do_entry_dirty", self.entry_dirty)
        c("on_addressbook_entry_cancel_clicked", self.do_cancel)
        c("on_addressbook_entry_delete_clicked", self.do_delete)
        c("on_addressbook_entry_ok_clicked", self.do_ok)
        c("on_grant_specific_toggled", self.do_specific_toggled)

        self.populateOwners()
        c("on_addressbook_owner_entry_changed", self.do_owner)

        owners = self.gui.agent.getOwners()
        if owners:
            self.setOwner(owners[0])
        self.do_select(sel)

    def populateOwners(self):
        owners = [owner.getName() for owner in self.gui.agent.getOwners()]
        if not owners:
            owners = [""]
        default = owners[0]
        self.w("addressbook_owner").set_popdown_strings(owners)
        self.w("addressbook_owner_entry").set_text(default)

    def do_owner(self, widget):
        # This is hooked to the GtkEntry's "changed" signal. When the entry
        # is changed by arrow-key-stepping through the GtkCombo, the
        # following signals are emitted:
        #  delete_text, changed, insert_text, changed
        # So we will be called twice, once while the Entry is empty.
        ownername = widget.get_text()
        if ownername != "":
            owner = self.gui.agent.getOwnerByName(ownername)
            self.setOwner(owner)

    def jumpToOwner(self, owner):
        self.w("addressbook_owner_entry").set_text(owner.getName())
        # it appears that this generates the right signal

    def setOwner(self, owner):
        if owner != self.owner and self.owner:
            oldbook = self.gui.agent.getAddressBook(self.owner.ID)
            oldbook.dontNotifyOnChange(self.updated)
        self.owner = owner

        book = self.gui.agent.getAddressBook(self.owner.ID)
        book.notifyOnChange(self.updated)

        self.populate()

    def do_lookup(self, widget):
        addr = self.w("lookup_entry").get_text()
        if addr:
            d = self.gui.agent.retrieveAddress(addr)
            d.addCallback(self._lookup_complete1)
            d.addErrback(log.err)
    def _lookup_complete1(self, signed):
        if not self.owner:
            return # weird
        book = self.gui.agent.getAddressBook(self.owner.ID)
        d = book.addFromSigned(signed)
        self.w("lookup_entry").set_text("")
        d.addCallback(self._lookup_complete2)
        return d

    def _lookup_complete2(self, entry):
        # try to select that entry, will do a self.populate()
        self.go_to_entry(entry)

    def populate(self):
        book = self.gui.agent.getAddressBook(self.owner.ID)
        self.addressbook_list.clear()
        for e in book.entries:
            iter = self.addressbook_list.append()
            self.addressbook_list.set(iter,
                                      0, e.getName(),
                                      1, e)
        #self.addressbook_tree.get_selection().select_path([0])
        self.addressbook_tree.get_selection().unselect_all()

    def go_to_entry(self, entry):
        """This selects both the owner and the entry."""
        owner = entry.addressbook.owner
        self.jumpToOwner(owner) # this populates the list
        model = self.addressbook_list
        for row in model:
            if row[1] == entry:
                self.addressbook_tree.scroll_to_cell(row.path)
                self.addressbook_tree.get_selection().select_path(row.path)
                return

    def do_specific_toggled(self, widget):
        specific_active = widget.get_active()
        disabled = ["IMT_toggle", "maxsize_toggle", "accept_html",
                    "grant_3rd_party"]
        self.set_sensitive(disabled, specific_active)
        IMT_active = self.w("IMT_toggle").get_active()
        if not specific_active:
            IMT_active = 0
        self.w("IMT_entry").set_sensitive(IMT_active)
        maxsize_active = self.w("maxsize_toggle").get_active()
        if not specific_active:
            maxsize_active = 0
        self.w("maxsize_entry").set_sensitive(maxsize_active)
        self.w("no_return_permission").set_active(0)

    def do_maxsize_toggled(self, widget):
        self.w("maxsize_entry").set_sensitive(widget.get_active())
    def do_IMG_toggled(self, widget):
        self.w("IMT_entry").set_sensitive(widget.get_active())

    def do_select(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            self.select_none()
            return
        entry = model.get_value(iter, 1)
        r = entry.idRecord
        assert(r)
        self.currentEntry = entry
        entry.notifyOnChange(self.entry_changed)
        self.select(entry)

    def entry_changed(self, results):
        self.select(self.currentEntry)

    def select_none(self):
        if self.currentEntry:
            self.currentEntry.dontNotifyOnChange(self.entry_changed)
        self.currentEntry = None
        self.w("ID_label").set_text("")
        self.w("addresses_label").set_text("")
        self.w("name_label").set_text("")
        self.w("petname_entry").set_text("")
        #self.w("IMT_entry")
        #self.w("maxsize_entry")
        #self.w("accept_html")
        self.w("outbound_permission_type").set_text(\
            "Their Restrictions on our Outbound")
        self.w("outbound_IMT").set_text("")
        self.w("outbound_maxsize").set_text("")
        self.w("outbound_html").set_text("")
        self.w("outbound_ticket_urls_label").set_text("")

        disabled = ["petname_entry",
                    "ID_label", "addresses_label", "name_label",
                    "transport_label",
                    "vouch_for", "no_return_permission",
                    "grant_specific",
                    "IMT_toggle", "IMT_entry",
                    "maxsize_toggle", "maxsize_entry",
                    "accept_html", "grant_3rd_party",
                    "addressbook_entry_delete"]
        self.set_sensitive(disabled, 0)

    def select(self, entry):
        r = entry.idRecord
        # names and such
        fields = ["petname_entry",
                  "ID_label", "addresses_label", "name_label",
                  "transport_label",
                  "vouch_for", "no_return_permission",
                  "grant_specific"]
        self.set_sensitive(fields, 1)

        self.w("ID_label").set_text(entry.idRecord.ID)
        self.w("addresses_label").set_text(", ".join(r.addresses))
        if r.name != None:
            self.w("name_label").set_text(r.name)
        else:
            self.w("name_label").set_text("")
        w = self.w("petname_entry")
        if entry.petname != None:
            w.set_text(entry.petname)
        else:
            w.set_text("")

        t = entry.idRecord.transportString
        if t:
            self.w("transport_label").set_text(t)
        else:
            self.w("transport_label").set_text("<None>")

        if entry.vouchFor:
            self.w("vouch_for").set_active(1)
        else:
            self.w("vouch_for").set_active(0)

        if entry.doNotGrantReturnPermission:
            self.w("no_return_permission").set_active(1)
        else:
            self.w("no_return_permission").set_active(0)

        # inbound
        inbound = ["IMT_toggle", "maxsize_toggle", "accept_html",
                   "grant_3rd_party"]
        if entry.theirPermission:
            p = entry.theirPermission
            self.w("grant_specific").set_active(1)
            self.set_sensitive(inbound, 1)
        else:
            # display the genericPermission values read-only
            p = self.owner.genericPermission
            self.w("grant_specific").set_active(0)
            self.set_sensitive(inbound, 0)

        if p.minIMG != None:
            self.w("IMT_toggle").set_active(1)
            self.w("IMT_entry").set_value(p.minIMG)
            self.w("IMT_entry").set_sensitive(1)
        else:
            self.w("IMT_toggle").set_active(0)
            self.w("IMT_entry").set_sensitive(0)

        if p.maxsize != None:
            self.w("maxsize_toggle").set_active(1)
            maxsize = float(p.maxsize / 1000)
            self.w("maxsize_entry").set_value(maxsize)
            self.w("maxsize_entry").set_sensitive(1)
        else:
            self.w("maxsize_toggle").set_active(0)
            self.w("maxsize_entry").set_sensitive(0)

        if p.contentTypes == ["*/*"]: # TODO: needs work
            self.w("accept_html").set_active(1)
        else:
            self.w("accept_html").set_active(0)

        if p.canGrantPermission:
            self.w("grant_3rd_party").set_active(1)
        else:
            self.w("grant_3rd_party").set_active(0)

        if not entry.theirPermission:
            self.set_sensitive(["IMT_entry", "maxsize_entry"], 0)

        # outbound
        p = entry.ourPermission
        if p:
            self.w("outbound_permission_type").set_text(\
                "Their Restrictions on our Outbound: specific")
        elif entry.idRecord.genericPermission:
            p = entry.idRecord.genericPermission
            self.w("outbound_permission_type").set_text(\
                "Their Restrictions on our Outbound: generic")
        else:
            self.w("outbound_permission_type").set_text(\
                "They do not accept inbound messages.")

        if p:
            minIMG = p.minIMG
            if minIMG == None:
                self.w("outbound_IMT").set_text("unlimited")
            else:
                self.w("outbound_IMT").set_text("%d seconds" % minIMG)
            maxsize = p.maxsize
            if maxsize == None:
                self.w("outbound_maxsize").set_text("unlimited")
            else:
                self.w("outbound_maxsize").set_text("%d kb" % \
                                                    (maxsize / 1000))
            if p.contentTypes == ["*/*"]: # TODO: needs work
                self.w("outbound_html").set_text("Accepts HTML mail")
            else:
                self.w("outbound_html").set_text("Does NOT accept HTML mail")
            urls = ", ".join(p.ticketURLs)
            self.w("outbound_ticket_urls_label").set_text(urls)
            # we do not bother to display .canGrantPermission

        self.set_sensitive(["addressbook_entry_cancel",
                            "addressbook_entry_ok"], 0)
        self.set_sensitive(["addressbook_entry_delete"], 1)

    def petname_entry_changed(self, widget):
        new_petname = widget.get_text()
        if self.currentEntry:
            if new_petname:
                # TODO: insure its uniqueness
                self.currentEntry.petname = new_petname
            else:
                self.currentEntry.petname = None
            # update the list
            name = self.currentEntry.petname
            if name == None:
                name = self.currentEntry.idRecord.ID
            sel = self.addressbook_tree.get_selection()
            model, iter = sel.get_selected()
            model.set(iter, 0, self.currentEntry.getName())
            self.gui.scheduleSave()

    def entry_dirty(self, *args):
        self.set_sensitive(["addressbook_entry_cancel",
                            "addressbook_entry_ok"], 1)

    def do_cancel(self, widget):
        self.select(self.currentEntry)

    def do_ok(self, widget):
        entry = self.currentEntry

        entry.vouchFor = self.w("vouch_for").get_active()
        no_return_permission = self.w("no_return_permission").get_active()
        entry.doNotGrantReturnPermission = no_return_permission

        if self.w("grant_specific").get_active():
            if not entry.theirPermission:
                # TODO: replace with agent.grantInboundPermission ??
                p = permission.OwnerPermission()
                entry.setTheirPermission(p)
            else:
                p = entry.theirPermission

            # set fields
            if self.w("IMT_toggle").get_active():
                IMG = self.w("IMT_entry").get_value()
            else:
                IMG = None
            p.setMinIMG(IMG)
            if self.w("maxsize_toggle").get_active():
                maxsize = int(self.w("maxsize_entry").get_value()) * 1000
                p.setMaxSize(maxsize)
            else:
                p.setMaxSize(None)
            if self.w("accept_html").get_active():
                p.setContentTypes(["*/*"])
            else:
                p.setContentTypes(["text/plain"]) # TODO: needs work
            if self.w("grant_3rd_party").get_active():
                p.canGrantPermission = True
            else:
                p.canGrantPermission = False

        else:
            if entry.theirPermission:
                entry.revokeTheirPermission()

        self.gui.agent.log("marking entry as modified")
        entry.modified() # this triggers a re-select


    def do_delete(self, widget):
        # confirm
        xml = gtk.glade.XML(gladefile, "confirm_dialog")
        dialog = xml.get_widget("confirm_dialog")
        l = xml.get_widget("confirm_label")
        confirm = ""
        entry = self.currentEntry
        if not entry:
            return
        if entry.theirPermission:
            confirm += "Their specific Permission will be revoked.\n"
        confirm += "\n"
        confirm += "<b>Really delete entry '%s'?</b>\n" % entry.getName()
        l.set_markup(confirm)
        dialog.set_title("Delete Entry %s" % entry.getName())
        dialog.connect("response", self._confirm_response, entry)
    def _confirm_response(self, dialog, responseid, entry):
        if responseid == gtk.RESPONSE_OK:
            self.really_delete(entry)
        dialog.destroy()
    def really_delete(self, entry):
        book = self.gui.agent.getAddressBook(self.owner.ID)
        book.deleteEntry(entry)

    def updated(self, addrbook):
        self.populate()
        self.gui.scheduleSave()

class RefsPane(Helper):
    def __init__(self, sendpane):
        xml = gtk.glade.XML(gladefile, "addrefs")
        self.xml = xml
        self.parent = sendpane
        self.window = xml.get_widget("addrefs")
        self.window.connect("destroy", self.do_destroy)
        # activate doesn't work with combo boxes, it seems
        #xml.signal_connect("on_refs_addr_activate", self.do_add)
        xml.signal_connect("do_addrefs_close", self.do_close)
        xml.signal_connect("do_add_ref", self.do_add)
        self.populate()
        self.updateRefs()

    def show(self):
        self.window.deiconify()

    def do_destroy(self, widget):
        self.parent.addrefs_closed()

    def do_add(self, widget):
        name = self.w("refs_adder_entry").get_text().strip()
        ownername = self.parent.w("send_frommenu_entry").get_text()
        owner = self.parent.gui.agent.getOwnerByName(ownername)
        book = self.parent.gui.agent.getAddressBook(owner.ID)
        r = book.lookupPetName(name)
        if not r:
            r = book.lookupAddress(name)
        if r:
            self.addRef(r)

    def do_close(self, widget):
        self.window.destroy()

    def addRef(self, entry):
        self.parent.refs.append(entry)
        self.updateRefs()

    def updateRefs(self):
        for c in self.w("refs_box").get_children():
            c.destroy()
        for entry in self.parent.refs:
            box = gtk.HBox()
            box.pack_start(gtk.Label(entry.getName()))
            b = gtk.Button("View")
            b.connect("clicked", self.do_view, entry)
            box.pack_end(b, expand=False)
            b = gtk.Button("Remove")
            b.connect("clicked", self.do_remove, entry)
            box.pack_end(b, expand=False)
            box.show_all()
            self.w("refs_box").add(box)

    def do_view(self, widget, entry):
        addrpane = self.parent.gui.addrpane
        addrpane.go_to_entry(entry)
        self.parent.gui.xml.get_widget("notebook").set_current_page(1)

    def do_remove(self, widget, entry):
        self.parent.refs.remove(entry)
        self.updateRefs()

    def populate(self):
        ownername = self.parent.w("send_frommenu_entry").get_text()
        owner = self.parent.gui.agent.getOwnerByName(ownername)
        book = self.parent.gui.agent.getAddressBook(owner.ID)
        addrs = []
        for e in book.entries:
            addrs.append(e.getName())
        if not addrs:
            addrs = [""]
        default = addrs[0]
        self.w("refs_adder").set_popdown_strings(addrs)
        self.w("refs_adder_entry").set_text(default)


class SendPane(Helper):
    def __init__(self, gui):
        self.gui = gui
        self.refpane = None
        self.refs = []

    def load(self, xml):
        self.xml = xml
        xml.signal_connect("do_send", self.do_send)
        xml.signal_connect("do_addref", self.do_addref)
        xml.signal_connect("do_clear", self.do_clear)
        self.recipientEntry = None
        xml.signal_connect("on_to_entry_changed", self.do_to_changed)
        self.populateOwners()

    def do_to_changed(self, widget):
        self.recipientEntry = None
        self.refs = []
        if self.refpane:
            self.refpane.window.destroy()
        return False

    def populateOwners(self):
        owners = [owner.getName() for owner in self.gui.agent.getOwners()]
        owners = ["-anonymous-"] + owners
        if len(owners) > 1:
            default = owners[1]
        else:
            default = owners[0]
        self.w("send_frommenu").set_popdown_strings(owners)
        self.w("send_frommenu_entry").set_text(default)

    def do_clear(self, button):
        self.w('to_entry').set_text("")
        self.w('subject_entry').set_text("")
        self.w('send_body').get_buffer().set_text("")
        self.refs = []
        if self.refpane:
            self.refpane.window.destroy()

    def do_addref(self, button):
        if not self.refpane:
            self.refpane = RefsPane(self)
        self.refpane.show() # make sure it is on top and not iconified

    def addrefs_closed(self):
        self.refpane = None

    def do_send(self, button):
        print "sending"
        ownername = self.w("send_frommenu_entry").get_text()
        if ownername == "-anonymous-":
            owner = None
        else:
            owner = self.gui.agent.getOwnerByName(ownername)
        if not owner:
            print "Hey, no owner.. anonymous sending not yet implemented"
            return

        m = message.Message(self.gui.agent)
        m.setSenderEntry(owner)

        if self.recipientEntry:
            m.setRecipientEntry(self.recipientEntry)
        else:
            name = self.w('to_entry').get_text().strip()
            m.setMessageRecipient(name, owner)

        subject = self.w('subject_entry').get_text()
        buf = self.w('send_body').get_buffer()
        start, end = buf.get_bounds()
        body = buf.get_text(start, end)
        fullbody = "From: %s\n" % owner.names + \
                   "To: you\n" + \
                   "Subject: %s\n" % subject + \
                   "\n" + body
        if fullbody[-1] != "\n":
            fullbody += "\n"
        m.addRFC822(fullbody)

        for e in self.refs:
            m.addReference(e)

        m.doDump = True
        m.send()

FROM = 0
SUBJECT = 1
MESSAGE = 2
class InboxPane(Helper):

    def __init__(self, gui):
        self.gui = gui
        self.currentMessage = None

    def load(self, xml):
        self.xml = xml
        model = self.messages = gtk.ListStore(gobject.TYPE_STRING, # from
                                              gobject.TYPE_STRING, # subject
                                              gobject.TYPE_PYOBJECT, # msg
                                              )
        view = self.w("inbox_tree")
        view.set_model(model)
        view.append_column(gtk.TreeViewColumn("From",
                                              gtk.CellRendererText(),
                                              text=FROM))
        view.append_column(gtk.TreeViewColumn("Subject",
                                              gtk.CellRendererText(),
                                              text=SUBJECT))
        sel = view.get_selection()
        sel.set_mode(gtk.SELECTION_SINGLE)
        sel.connect("changed", self.do_message)

        xml.signal_connect("on_inbox_delete_clicked", self.do_delete)
        xml.signal_connect("on_inbox_reply_clicked", self.do_reply)

    def receive(self, msg, why):
        print "gui received"
        fromName = None
        msg.why = why
        if msg.senderEntry:
            fromName = msg.senderEntry.getName()
        if fromName is None:
            fromName = "??"
        self.messages.append((fromName, "<notyet>", msg))
        msg.body = msg.getNode('rfc822')['body']
        msg.doDump = True
        self.gui.scheduleSave()

    def do_message(self, sel):
        model, iter = sel.get_selected()
        if not iter:
            return
        msg = model.get_value(iter, MESSAGE)
        self.currentMessage = msg
        self.populate()


    def addButton(self, label, text):
        box = gtk.HBox()
        b = gtk.Button(label)
        box.pack_start(b, expand=False)
        box.pack_end(gtk.Label(text))
        box.show_all()
        self.w("inbox_buttons").add(box)
        return b

    def populate(self):
        msg = self.currentMessage
        self.w('from_label').set_text(msg.getFromName())
        self.w('from_permission_label').set_text(msg.why)
        self.w('inbox_body').get_buffer().set_text(msg.body)

        self.populate_buttons()

    def populate_buttons(self):
        msg = self.currentMessage
        buttons = self.w("inbox_buttons")
        # clear the buttonbox
        for c in buttons.get_children():
            c.destroy()
        if msg.canAddSender():
            b = self.addButton("Add Sender",
                               "Add %s to your Addressbook" % \
                               msg.getFromName())
            b.connect("clicked", self.do_add_sender, msg)
        if msg.canReply():
            b = self.addButton("Reply",
                               "Reply to this message")
            b.connect("clicked", self.do_reply)
        for r in msg.canAddReferences():
            b = self.addButton("Add Reference",
                               "Add %s to your Addressbook" % r['name'])
            b.connect("clicked", self.do_add_reference, r, msg)

        # add from-buttons
        for c in self.w('inbox_from_buttons').get_children():
            c.destroy()
        if msg.senderEntry:
            b = gtk.Button("View")
            self.w('inbox_from_buttons').add(b)
            b.connect("clicked", self.do_view_sender, msg)
            b.show()


    def do_add_sender(self, widget, msg):
        d = msg.addSender()
        d.addCallback(self._do_add_sender1)
        self.populate_buttons()

    def _do_add_sender1(self, entry):
        # make the address-pane jump to the entry, but don't actually switch
        # to that pane
        self.gui.addrpane.go_to_entry(entry)
        self.populate_buttons()

    def do_add_reference(self, widget, refnode, msg):
        d = msg.addReference(refnode)
        d.addCallback(self._do_add_sender1)

    def do_view_sender(self, widget, msg):
        addrpane = self.gui.addrpane
        addrpane.go_to_entry(msg.senderEntry)
        self.gui.xml.get_widget("notebook").set_current_page(1)

    def do_delete(self, widget):
        print "Not implemented"

    def do_reply(self, widget):
        msg = self.currentMessage
        sendpane = self.gui.sendpane
        sender_name = msg.recipientEntry.getName()
        sendpane.w("send_frommenu_entry").set_text(sender_name)
        sendpane.w("to_entry").set_text(msg.getFromName())
        sendpane.recipientEntry = msg.senderEntry
        sendpane.w("subject_entry").set_text("Reply to your message")
        sendpane.w("send_body").get_buffer().set_text("")
        # TODO: also make the TextView active ("give it focus"?)
        self.gui.xml.get_widget("notebook").set_current_page(2)

class SheepBox(object):
    def __init__(self, gui):
        self.gui = gui
    def load(self, xml):
        self.area = xml.get_widget("sheeparea")
        self.label = xml.get_widget("sheeplabel")
        self.entry = xml.get_widget("sheepentry")
        self.plantSheep()
        self.showEntry(0)

    def OLDmakeLabel(self, text):
        self.label.set_text(text)
        if self.boxContents == self.label:
            return
        if self.boxContents != None:
            self.boxContents.hide()
            self.box.remove(self.boxContents)
        self.box.add(self.label)
        self.boxContents = self.label
        self.boxContents.show_all()
    def OLDmakeEntry(self, instructions):
        self.vboxInstructions.set_text(instructions)
        if self.boxContents == self.vbox:
            return
        if self.boxContents != None:
            self.boxContents.hide()
            self.box.remove(self.boxContents)
        self.box.add(self.vbox)
        self.boxContents = self.vbox
        self.boxContents.show_all()

    def showEntry(self, show):
        if show:
            self.entry.show()
        else:
            self.entry.hide()

    def plantSheep(self):
        self.area.set_from_file("docs/sheep.gif")

class SetupPane(Helper):
    def __init__(self, parent):
        self.parent = parent
    def loadXML(self):
        self.xml = xml = gtk.glade.XML(gladefile, "setup")
        self.window = window = xml.get_widget("setup")
        window.connect("destroy", self.do_destroy)
        self.ok_button = xml.get_widget("setup_ok_button")
        c = xml.signal_connect
        c("do_dirty", self.do_dirty)

        c("on_setup_address_pb_toggled", self.address_pb_toggled)
        c("on_setup_address_disk_toggled", self.address_disk_toggled)

        c("on_setup_mail_smtp_toggled", self.mail_smtp_toggled)
        c("on_setup_mail_mixmaster_toggled", self.mail_mixmaster_toggled)
        c("on_setup_mail_mixminion_toggled", self.mail_mixminion_toggled)
        
        c("on_setup_nntp_needpass_toggled", self.nntp_needpass_toggled)

        c("on_setup_close", self.do_close)
        c("on_setup_ok", self.do_ok)
        
        self.populate()
        self.ok_button.set_sensitive(0)

    def do_close(self, widget):
        self.window.destroy()

    def populate(self):
        agent = self.parent.agent

        lookup = agent.getAddressLookup()
        if lookup.startswith("PB:"):
            self.w("setup_address_pb").set_active(1)
            self.w("setup_address_disk").set_active(0)
            self.w("setup_address_fake").set_active(0)
            # TODO: use an instance instead of these strings
            scheme, host, port = lookup.split(":")
            self.w("setup_address_pb_host").set_text(host)
            self.w("setup_address_pb_port").set_text(port)
        elif lookup.startswith("Disk:"):
            self.w("setup_address_pb").set_active(0)
            self.w("setup_address_disk").set_active(1)
            self.w("setup_address_fake").set_active(0)
        elif lookup.startswith("Fake:"):
            self.w("setup_address_pb").set_active(0)
            self.w("setup_address_disk").set_active(0)
            self.w("setup_address_fake").set_active(1)

        mail = agent.smtpScheme
        if mail.startswith("Mail:"):
            self.w("setup_mail_smtp").set_active(1)
            scheme, relayhost, fromaddr = mail.split(":")
            self.w("setup_smtp_host").set_text(relayhost)
            self.w("setup_smtp_from").set_text(fromaddr)
        elif mail.startswith("Mixmaster:"):
            self.w("setup_mail_mixmaster").set_active(1)
        elif mail.startswith("Mixminion:"):
            self.w("setup_mail_mixminion").set_active(1)

        nntp = agent.nntpScheme
        if nntp:
            pass
        self.w("setup_nntp_needpass").set_active(0)

        # now emit "toggled" signals on all these widgets so they do their
        # update things
        for n in ["setup_address_pb",
                  "setup_address_disk",
                  "setup_address_fake",

                  "setup_mail_smtp",
                  "setup_mail_mixmaster",
                  "setup_mail_mixminion",

                  "setup_nntp_needpass"]:
            self.w(n).emit("toggled")
        

    def do_dirty(self, widget):
        self.ok_button.set_sensitive(1)

    def address_pb_toggled(self, widget):
        fields = ["setup_address_pb_host_label", "setup_address_pb_host",
                  "setup_address_pb_port_label", "setup_address_pb_port",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def address_disk_toggled(self, widget):
        fields = ["setup_address_disk_directory_label",
                  "setup_address_disk_directory",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def address_fake_toggled(self, widget):
        fields = ["setup_address_fake_label",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def mail_smtp_toggled(self, widget):
        fields = ["setup_smtp_host_label", "setup_smtp_host",
                  "setup_smtp_from_label", "setup_smtp_from"
                  ]
        self.set_sensitive(fields, widget.get_active())
    def mail_mixmaster_toggled(self, widget):
        fields = ["setup_mail_mixmaster_label"]
        self.set_sensitive(fields, widget.get_active())
    def mail_mixminion_toggled(self, widget):
        fields = ["setup_mail_mixminion_label"]
        self.set_sensitive(fields, widget.get_active())

    def nntp_needpass_toggled(self, widget):
        fields = ["setup_nntp_user_label", "setup_nntp_user",
                  "setup_nntp_pass_label", "setup_nntp_pass",
                  ]
        self.set_sensitive(fields, widget.get_active())

    def do_ok(self, widget):
        agent = self.parent.agent

        if self.w("setup_address_pb").get_active():
            lookup = "PB:%s:%s" % (self.text("setup_address_pb_host"),
                                   self.text("setup_address_pb_port"))
        elif self.w("setup_address_disk").get_active():
            lookup = "Disk:%s" % self.text("setup_address_disk_directory")
        elif self.w("setup_address_fake").get_active():
            lookup = "Fake:"
        else:
            print "Weird address lookup: nothing selected"
            
        agent.setAddressLookup(lookup)

        if self.w("setup_mail_smtp").get_active():
            w = self.w("setup_smtp_host")
            relayhost = w.get_text()
            w = self.w("setup_smtp_from")
            smtpfrom = w.get_text()
            smtp = "Mail:%s:%s" % (relayhost, smtpfrom)
            agent.setSMTP(smtp)
        else:
            print "WARNING: mixmaster/mixminion mail outlets not implemented"
            print "reverting to using SMTP for outbound mail"

        host = self.w("setup_nntp_host").get_text()
        port = self.w("setup_nntp_port").get_text()
        port = int(port)
        if self.w("setup_nntp_needpass").get_active():
            user = self.w("setup_nntp_user").get_text()
            passwd = self.w("setup_nntp_pass").get_text()
            nntp = "%s:%s:%s:%s" % (host, port, user, passwd)
        else:
            nntp = "%s:%s" % (host, port)
        agent.setNNTP(nntp)

        self.window.destroy()
        self.parent.scheduleSave()

    def show(self):
        self.window.deiconify()
    def do_destroy(self, widget):
        self.parent.closeSetup()

        
class Gui(object):
    def __init__(self, agent):
        self.agent = agent
        self.addrpane = AddressBookPane(self)
        self.mepane = MePane(self)
        self.sendpane = SendPane(self)
        self.inboxpane = InboxPane(self)
        self.sheepbox = SheepBox(self)
        self.logpane = None
        self.saveTimer = None
        self.setup = None

    def scheduleSave(self):
        if self.saveTimer:
            self.saveTimer.cancel()
        self.saveTimer = reactor.callLater(5, self.do_save)

    def do_quit(self, widget):
        if self.saveTimer:
            self.saveTimer.cancel()
            self.saveTimer = None
        self.do_save()
        self.agent.stop()

    def do_setup(self, widget):
        if self.setup:
            self.setup.show()
            return
        self.setup = SetupPane(self)
        self.setup.loadXML()
    def closeSetup(self):
        self.setup = None

    def do_event_log(self, widget):
        if self.logpane:
            self.logpane.show()
        else:
            self.logpane = LogWindow(self)
    def event_log_closed(self):
        print "log closed"
        self.logpane = None

    def do_debug_shell(self, widget):
        print "unimplemented"

    def loadXML(self):
        self.xml = xml = gtk.glade.XML(gladefile, "app")
        xml.signal_connect("do_setup", self.do_setup)
        xml.signal_connect("do_quit", self.do_quit)
        xml.signal_connect("do_event_log", self.do_event_log)
        xml.signal_connect("do_debug_shell", self.do_debug_shell)
        self.addrpane.load(xml)
        self.mepane.load(xml)
        self.sendpane.load(xml)
        self.inboxpane.load(xml)
        self.sheepbox.load(xml)
        self.ticketAgents = [TicketAgent(self)]

    def do_save(self):
        if self.saveTimer:
            self.saveTimer = None
        if self.savefile:
            print "saving to", self.savefile
            tmpfile = self.savefile + ".tmp"
            try:
                pickle.dump(self.agent, open(tmpfile, "w"))
                os.rename(tmpfile, self.savefile)
            except (OSError, IOError), x:
                print "couldn't save:", x

class GuiAgent(agent.Agent):
    def __getstate__(self):
        d = agent.Agent.__getstate__(self)
        del d['gui']
        return d

    def setupGui(self, gui):
        self.gui = gui

    def getTicketAgents(self):
        return self.gui.ticketAgents
    def acceptRFC822(self, rm, why):
        return self.gui.inboxpane.receive(rm, why)

def run(basedir):
    print "basedir", basedir
    savefile = os.path.join(basedir, "gui.pickle")
    if os.path.exists(savefile):
        print "loading"
        a = pickle.load(open(savefile, "r"))
    else:
        print "initializing"
        if not os.path.exists(basedir):
            os.mkdir(basedir)
        a = GuiAgent(basedir)
        a.init()
        #a.setAddressLookup("Disk:_DiskNameServer")

    log.addObserver(a.eventLog.addEvent)

    gui = Gui(a)
    a.setupGui(gui)
    gui.savefile = savefile
    gui.loadXML()
    gui.mepane.populate()
    a.run()

if __name__ == '__main__':
    import sys
    savefile = None
    if len(sys.argv) > 1:
        savefile = sys.argv[1]
    run(savefile)
